﻿// -*- C++ -*-
// -*- coding: utf-8 -*-
/*!
 * @file  TRobotRTC.hpp
 * @brief TRobot Component
 *
 */
#ifndef TROBOT_RTC_HPP
#define TROBOT_RTC_HPP

#include <rtm/idl/BasicDataTypeSkel.h>
#include <rtm/idl/ExtendedDataTypesSkel.h>
#include <rtm/Manager.h>
#include <rtm/DataFlowComponentBase.h>
#include <rtm/CorbaPort.h>
#include <rtm/DataInPort.h>
#include <rtm/DataOutPort.h>

#include "TRobotSerialIOIface.hpp"
#include "TRobotSensorData.hpp"

// Service implementation headers
// <rtc-template block="service_impl_h">

// </rtc-template>

// Service Consumer stub headers
// <rtc-template block="consumer_stub_h">

// </rtc-template>


////////////////////////////////////////////////////////////////////////////////
/**
 *  @brief Tロボットベース用RTコンポーネント
 *
 *  @copybrief
 *
 *  シリアルポート経由でロボットベースのメイン基板との通信（制御指令の送信、センサデータの受信）を行う。
 *
 *  RTC入力ポート:
 *  @li @c refVel 速度目標値 (RTC::TimedVelocity2D)
 *
 *  RTC出力ポート: 
 *  @li @c odometryPose オドメトリ位置姿勢データ (RTC::TimedPose2D)
 */
class TRobotRTC  : public RTC::DataFlowComponentBase
{
 public:
  TRobotRTC(RTC::Manager* manager);
  ~TRobotRTC();

  virtual RTC::ReturnCode_t onInitialize();
  virtual RTC::ReturnCode_t onActivated(RTC::UniqueId ecID);
  virtual RTC::ReturnCode_t onDeactivated(RTC::UniqueId ecID);
  virtual RTC::ReturnCode_t onExecute(RTC::UniqueId ecID);
  
  // virtual RTC::ReturnCode_t onFinalize();
  // virtual RTC::ReturnCode_t onStartup(RTC::UniqueId ec_id);
  // virtual RTC::ReturnCode_t onShutdown(RTC::UniqueId ec_id);
  // virtual RTC::ReturnCode_t onAborting(RTC::UniqueId ec_id);
  // virtual RTC::ReturnCode_t onError(RTC::UniqueId ec_id);
  // virtual RTC::ReturnCode_t onReset(RTC::UniqueId ec_id);
  // virtual RTC::ReturnCode_t onStateUpdate(RTC::UniqueId ec_id);
  // virtual RTC::ReturnCode_t onRateChanged(RTC::UniqueId ec_id);


protected:
  // Configuration variable declaration
  // <rtc-template block="config_declare">
  std::string m_serialPortName;  //!< シリアルポートの名称（Linux: /dev/ttyUSB0, etc.,  Windows: COM3, etc.）
  double m_wheelRadius;          //!< 駆動輪の半径 [m]
  double m_axleTrack;            //!< 輪距（左右の駆動輪中心間の距離） [m]
  double m_rotEncoderResolution; //!< 駆動輪モータ軸のエンコーダの分解能 [counts/rev]
  double m_wheelGearRatio;       //!< 駆動輪の減速機の減速比 [無単位]
  // </rtc-template>
  


  // DataInPort declaration
  // <rtc-template block="inport_declare">
  RTC::InPort<RTC::TimedVelocity2D> m_refVelIn; //!< 速度目標値用RTC入力ポート
  // </rtc-template>

  
  // DataOutPort declaration
  // <rtc-template block="outport_declare">
  RTC::OutPort<RTC::TimedPose2D> m_odometryPoseOut; //!< オドメトリ位置姿勢データ用RTC出力ポート
  // </rtc-template>

  // CORBA Port declaration
  // <rtc-template block="corbaport_declare">

  // </rtc-template>

  // Service declaration
  // <rtc-template block="service_declare">

  // </rtc-template>

  // Consumer declaration
  // <rtc-template block="consumer_declare">

  // </rtc-template>
  

 private:
  void updateOdometry(TRobot::BaseSensorData& sensorData, RTC::Pose2D& odometryPose);
  void resetOdometry(RTC::Pose2D& odometryPose);
  TRobot::RetCode convertToWheelRotVelCmds(double baseFwdVel,  double baseAngVel,
                                           int& leftRotVelCmd, int& rightRotVelCmd);
  
  TRobot::SerialIOIface *m_serIO;      //!< シリアル入出力用インスタンスへのポインタ
  TRobot::BaseSensorData m_sensorData; //!< ロボットベースのメイン基板に接続されたセンサのデータ（エンコーダ値, etc.）

  RTC::TimedVelocity2D m_refVel;   //!< RTC入力ポートから入力される速度目標値
  RTC::TimedPose2D m_odometryPose; //!< オドメトリ位置姿勢データ

  
  /*!
   * オドメトリ用変換係数。
   * エンコーダカウント値から，対応する回転により描く弧長の半分の値への変換
   * (その値はコンフィグパラメータに依存するので onActivated() で計算される)
   */
  double m_coeffEncToArcHalf;

  double m_coeffRpsToCmd; //!< 角速度 [rad/s] から回転速度指令値への変換係数
  
  double m_halfAxleTrack; //!< 輪距（左右の駆動輪中心間の距離）の半分 [m]
};


extern "C"
{
  DLL_EXPORT void TRobotRTCInit(RTC::Manager* manager);
};

#endif // TROBOT_RTC_HPP
