﻿// -*- C++ -*-
// -*- coding: utf-8 -*- 
/*!
 *  @file TRobotRTC.cpp
 *  @brief TRobotRTCの実装
 *
 */


#include <cmath> 

#ifdef _WIN32
#  include "TRobotSerialIOWindows.hpp"
#  define DEFAULT_SERIAL_PORT_NAME "COM13" //!< TRobotRTC::m_serialPortName のデフォルト値
#else
#  include "TRobotSerialIOLinux.hpp"
#  define DEFAULT_SERIAL_PORT_NAME "/dev/ttyACM0" //!< TRobotRTC::m_serialPortName のデフォルト値 
#endif


#include "TRobotRTC.hpp"
#include "TRobotUtil.hpp"



#define DEFAULT_WHEEL_RADIUS "0.062" //!< TRobotRTC::m_wheelRadius のデフォルト値
#define DEFAULT_AXLE_TRACK   "0.400" //!< TRobotRTC::m_axleTrack のデフォルト値 
#define DEFAULT_ROT_ENCODER_RESOLUTION "800" //!< TRobotRTC::m_rotEncoderResolution のデフォルト値
#define DEFAULT_WHEEL_GEAR_RATIO "15" //!< TRobotRTC::m_wheelGearRatio のデフォルト値


// Module specification
static const char* trobotrtc_spec[] = {
  "implementation_id", "TRobotRTC",
  "type_name",         "TRobotRTC",
  "description",       "TRobot Component",
  "version",           "1.0",
  "vendor",            "Not available",
  "category",          "VehicleDriver",
  "activity_type",     "SPORADIC",
  "kind",              "DataFlowComponent",
  "max_instance",      "10",
  "language",          "C++",
  "lang_type",         "compile",
  // Configuration variables
  "conf.default.serialPortName",   DEFAULT_SERIAL_PORT_NAME,
  "conf.default.wheelRadius",      DEFAULT_WHEEL_RADIUS,
  "conf.default.axleTrack",        DEFAULT_AXLE_TRACK,
  "conf.default.rotEncoderResolution", DEFAULT_ROT_ENCODER_RESOLUTION,
  "conf.default.wheelGearRatio",       DEFAULT_WHEEL_GEAR_RATIO,
  ""
};

////////////////////////////////////////////////////////////////////////////////
/*!
 *  @brief コンストラクタ
 *
 *  @copybrief
 *
 *  @param manager [in] RTCマネージャのインスタンスへのポインタ
 */
TRobotRTC::TRobotRTC(RTC::Manager* manager)
    // <rtc-template block="initializer">
  : RTC::DataFlowComponentBase(manager),
    m_refVelIn("refVel", m_refVel),
    m_odometryPoseOut("odometryPose", m_odometryPose)
    // </rtc-template>
{
#ifdef _WIN32
  m_serIO = new TRobot::SerialIOWindows();
#else  
  m_serIO = new TRobot::SerialIOLinux();
#endif  
}


////////////////////////////////////////////////////////////////////////////////
/*!
 *  @brief デストラクタ
 *
 *  @copybrief
 *
 */
TRobotRTC::~TRobotRTC()
{
  delete m_serIO;
}


////////////////////////////////////////////////////////////////////////////////
/*!
 *
 */
RTC::ReturnCode_t TRobotRTC::onInitialize()
{
  using namespace std;
  cout << "TRobotRTC::onInitialize()" << endl;
  
  // Registration: InPort/OutPort/Service
  // <rtc-template block="registration">
  // Set InPort buffers
  addInPort(m_refVelIn.getName(), m_refVelIn);
  
  // Set OutPort buffer
  addOutPort(m_odometryPoseOut.getName(), m_odometryPoseOut);
  

  // Set service provider to Ports

  // Set service consumers to Ports

  // Set CORBA Service Ports

  // </rtc-template>

  // <rtc-template block="bind_config">
  // Bind variables and configuration variable
  bindParameter("serialPortName", m_serialPortName, DEFAULT_SERIAL_PORT_NAME);
  bindParameter("wheelRadius",    m_wheelRadius,    DEFAULT_WHEEL_RADIUS);
  bindParameter("axleTrack",      m_axleTrack,      DEFAULT_AXLE_TRACK);
  bindParameter("rotEncoderResolution", m_rotEncoderResolution,
                DEFAULT_ROT_ENCODER_RESOLUTION);
  bindParameter("wheelGearRatio", m_wheelGearRatio, DEFAULT_WHEEL_GEAR_RATIO);
  // </rtc-template>
  
  return RTC::RTC_OK;
}


/*
  RTC::ReturnCode_t TRobotRTC::onFinalize()
  {
  return RTC::RTC_OK;
  }
*/
/*
  RTC::ReturnCode_t TRobotRTC::onStartup(RTC::UniqueId ec_id)
  {
  return RTC::RTC_OK;
  }
*/
/*
RTC::ReturnCode_t TRobotRTC::onShutdown(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/


////////////////////////////////////////////////////////////////////////////////
/*!
 *
 */
RTC::ReturnCode_t TRobotRTC::onActivated(RTC::UniqueId ecID)
{
  using namespace std;
  TRobot::RetCode retVal;

  // onExecute() で繰り返し計算しないようにするため，ここで計算する
  m_halfAxleTrack = m_axleTrack * 0.5;
  
  m_coeffEncToArcHalf = m_wheelRadius * M_PI / (m_rotEncoderResolution * m_wheelGearRatio);

  const double KMPH_TO_CMD = 32.0; // [km/h] から (直進時)指令値単位への変換係数
  const double KMPH_TO_MPS = (60.0 * 60.0) / 1000.0; // [m/s] から [km/h] への変換係数
  // [rad/s] からファームウェアの指令値単位への変換係数
  m_coeffRpsToCmd = m_wheelRadius * KMPH_TO_CMD * KMPH_TO_MPS;

  // オドメトリデータのリセット
  resetOdometry(m_odometryPose.data);

  // RTC入力ポートにデータが残っていない状態にする(遅延対策)
  // cout << "Deleting remaining data in input ports...";
  while(!m_refVelIn.isEmpty()){
    // cout << "." << flush;
    m_refVelIn.read();
  }
  // cout << endl;
    
  retVal = m_serIO->start(m_serialPortName.c_str());
  if(retVal != TRobot::SUCCESS){
    cerr << "Failed to start communication on " << m_serialPortName << endl;
    return RTC::RTC_ERROR;
  }
  return RTC::RTC_OK;
}


////////////////////////////////////////////////////////////////////////////////
/*!
 *
 */
RTC::ReturnCode_t TRobotRTC::onDeactivated(RTC::UniqueId ecID)
{
  using namespace std;
  TRobot::RetCode retVal;
  
  m_serIO->sendWheelRotVel(0, 0); // 安全対策として停止させる
  while(m_serIO->receiveSensorData(m_sensorData) == TRobot::SUCCESS){
    // 受信済みのセンサデータが尽きるまで読み捨てる
  }
  retVal = m_serIO->stop();
  if(retVal != TRobot::SUCCESS){
    cerr << "Failed to stop communication" << endl;
  }
  return RTC::RTC_OK;
}


////////////////////////////////////////////////////////////////////////////////
/*!
 *
 */
RTC::ReturnCode_t TRobotRTC::onExecute(RTC::UniqueId ecID)
{
  using namespace std;
  static int leftRotVelCmd = 0;
  static int rightRotVelCmd = 0;
  TRobot::RetCode retVal;

  // バッファに溜まったrefVelをすべて読み出す
  bool refVelUpdated = false;
  while(m_refVelIn.isNew()){
    m_refVelIn.read();
    refVelUpdated = true;
  }
  
  if(refVelUpdated){
    retVal = convertToWheelRotVelCmds(m_refVel.data.vx, m_refVel.data.va,
                                      leftRotVelCmd, rightRotVelCmd);
    if(retVal != TRobot::SUCCESS){
      cerr << "Error on Vel2D to rot vel conversion" << endl;
    }else{
      // cout << "(lrv, rrv) == (" << leftRotVel << ", " << rightRotVel << ")" << endl;
      m_serIO->sendWheelRotVel(leftRotVelCmd, rightRotVelCmd);
    }
  }else{
    // 直前の refVelIn に対応する速度指令を使う
    m_serIO->sendWheelRotVel(leftRotVelCmd, rightRotVelCmd);
  }
    
  // センサデータがキューにあればオドメトリ計算を行う
  while(m_serIO->receiveSensorData(m_sensorData) == TRobot::SUCCESS){
    updateOdometry(m_sensorData, m_odometryPose.data);
  }

  setTimestamp(m_odometryPose);
  m_odometryPoseOut.write();
  return RTC::RTC_OK;
}


////////////////////////////////////////////////////////////////////////////////
/*!
 *  @brief オドメトリデータを更新する
 *
 *  @copybrief
 *
 *  @param sensorData   [in]    最新のセンサデータ
 *  @param odometryPose [inout] 現在のオドメトリデータ
 */
void TRobotRTC::updateOdometry(TRobot::BaseSensorData& sensorData, RTC::Pose2D& odometryPose)
{
  for(size_t i = 0; i < sensorData.NUM_ENC_COUNTS; i++){
    int renc = sensorData.m_rightEnc[i];
    int lenc = sensorData.m_leftEnc[i];
    int sumEnc  = renc + lenc;
    double h = odometryPose.heading; // 方向角の一時保持用

#  if 0
    int diffEnc = lenc - renc;
#  else
    int diffEnc = renc - lenc;
#  endif
    const int DIFF_ENC_EPS = 5;
    
    if(abs(diffEnc) > DIFF_ENC_EPS){ // （旋回半径が大きいとしても）旋回している場合
      double rho = (m_halfAxleTrack * sumEnc) / diffEnc; // 符号付き瞬時旋回半径 [m]
      odometryPose.heading    += m_coeffEncToArcHalf * diffEnc / m_halfAxleTrack;
      odometryPose.position.x += rho * (-sin(h) + sin(odometryPose.heading));
      odometryPose.position.y += rho * (+cos(h) - cos(odometryPose.heading));
    }else{ // 直進している場合
      double dfwd = m_coeffEncToArcHalf*sumEnc; // 直進距離(or 後退距離) [m]
      odometryPose.position.x += dfwd * cos(h);
      odometryPose.position.y += dfwd * sin(h);
    }
  }
}


////////////////////////////////////////////////////////////////////////////////
/*!
 *  
 */
void TRobotRTC::resetOdometry(RTC::Pose2D& odometryPose)
{
  m_odometryPose.data.position.x = 0.0;
  m_odometryPose.data.position.y = 0.0;
  m_odometryPose.data.heading    = 0.0;
}


////////////////////////////////////////////////////////////////////////////////
/*!
 *  @brief ロボットベースの前進速度および旋回速度の組み合わせから
 *         左右の車輪の回転速度指令値への変換を行う
 *
 *  @copybrief
 *
 *  @param baseFwdVel [in] ロボットベースの前進速度
 *  @param baseAngVel [in] ロボットベースの旋回速度
 *  @param leftRotVelCmd  [out] 左車輪の回転速度指令値
 *  @param rightRotVelCmd [out] 右車輪の回転速度指令値
 */
TRobot::RetCode TRobotRTC::convertToWheelRotVelCmds(double baseFwdVel,
                                                    double baseAngVel,
                                                    int& leftRotVelCmd,
                                                    int& rightRotVelCmd)
{
  // 前提:
  //  * 上から見て反時計回りが正方向の旋回
  //  * スリップなし
  double halfATAngVel = m_halfAxleTrack * baseAngVel;
  double wl = (baseFwdVel - halfATAngVel) / m_wheelRadius;
  double wr = (baseFwdVel + halfATAngVel) / m_wheelRadius;
    
  leftRotVelCmd  = TRobot::limitRange((int)(m_coeffRpsToCmd * wl),
                                      TRobot::ROT_VEL_CMD_MIN,
                                      TRobot::ROT_VEL_CMD_MAX);
  
  rightRotVelCmd = TRobot::limitRange((int)(m_coeffRpsToCmd * wr),
                                      TRobot::ROT_VEL_CMD_MIN,
                                      TRobot::ROT_VEL_CMD_MAX);
    
  return TRobot::SUCCESS;  
}


/*
  RTC::ReturnCode_t TRobotRTC::onAborting(RTC::UniqueId ec_id)
  {
  return RTC::RTC_OK;
  }
*/
/*
  RTC::ReturnCode_t TRobotRTC::onError(RTC::UniqueId ec_id)
  {
  return RTC::RTC_OK;
  }
*/
/*
  RTC::ReturnCode_t TRobotRTC::onReset(RTC::UniqueId ec_id)
  {
  return RTC::RTC_OK;
  }
*/
/*
RTC::ReturnCode_t TRobotRTC::onStateUpdate(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/
/*
RTC::ReturnCode_t TRobotRTC::onRateChanged(RTC::UniqueId ec_id)
{
  return RTC::RTC_OK;
}
*/


extern "C"
{
  void TRobotRTCInit(RTC::Manager* manager)
  {
    coil::Properties profile(trobotrtc_spec);
    manager->registerFactory(profile,
                             RTC::Create<TRobotRTC>,
                             RTC::Delete<TRobotRTC>);
  }
};
