﻿#include "TRobotSerialIOLinux.hpp"
#include "TRobotSensorData.hpp"
#include "TRobotUtil.hpp"

#include <sys/types.h>
#include <sys/stat.h>
#include <fcntl.h>
#include <termios.h>
#include <unistd.h>
#include <pthread.h>

#include <iostream>
#include <algorithm>
#include <cstdio>
#include <cstring>


namespace
{
  ////////////////////////////////////////////////////////////////////////////////
  void printPosixError(const char *msg, int errnum)
  {
    const size_t MSG_BUF_LEN = 64;
    char *buf = new char[MSG_BUF_LEN];
    std::cerr << msg << ": "
              << strerror_r(errnum, buf, MSG_BUF_LEN)
              << std::endl;
    delete buf;  
  }

  ////////////////////////////////////////////////////////////////////////////////
  void initMutexErrCheck(pthread_mutex_t *mutex)
  {
    int retVal = pthread_mutex_init(mutex, NULL);
    if(retVal != 0){
      printPosixError("Failed to initialize a mutex", retVal);
    }
  }

  ////////////////////////////////////////////////////////////////////////////////
  void destroyMutexErrCheck(pthread_mutex_t *mutex)
  {
    int retVal = pthread_mutex_destroy(mutex); 
    if(retVal != 0){
      printPosixError("Failed to destroy a mutex", retVal);
    }
  }

  ////////////////////////////////////////////////////////////////////////////////
  void joinThreadErrCheck(pthread_t *threadId, const char *threadName)
  {
    int retVal;
    int *thrRetVal;
        
    retVal = pthread_join(*threadId, (void **)&thrRetVal);
    if(retVal != 0){
      printPosixError("Faild to join a thread", retVal);
    }
    std::cout << "The " << threadName << " thread exited with status code "
              << thrRetVal << std::endl;
  }
}

////////////////////////////////////////////////////////////////////////////////
/*!
 *  コンストラクタ
 */
TRobot::SerialIOLinux::SerialIOLinux() :
  m_stopRequested(false),
  m_isRunning(false)
{
  initMutexErrCheck(&m_stopReqLock);
  initMutexErrCheck(&m_senDataQLock);
  initMutexErrCheck(&m_rotVelCmdQLock);
    
  int retVal = pthread_cond_init(&m_condFirstSendDone, NULL);
  if(retVal != 0){
    printPosixError("Failed to initialize a condition variable", retVal);
  }
}


////////////////////////////////////////////////////////////////////////////////
/*!
 *  デストラクタ
 */
TRobot::SerialIOLinux::~SerialIOLinux()
{
  destroyMutexErrCheck(&m_stopReqLock);   
  destroyMutexErrCheck(&m_senDataQLock);  
  destroyMutexErrCheck(&m_rotVelCmdQLock);

  int retVal = pthread_cond_destroy(&m_condFirstSendDone);
  if(retVal != 0){
    printPosixError("Failed to destroy a condition variable", retVal);
  }
}


////////////////////////////////////////////////////////////////////////////////
/*!
 *  @brief ロボットベースのファームウェアとの通信を開始する
 *
 *  @copybrief
 *
 *  @param port 通信ポートを表す文字列(/dev/ttyUSB0 など) 
 */
TRobot::RetCode TRobot::SerialIOLinux::start(const char *port)
{
  using namespace std;
    
  struct termios termCfg;
  int retVal;
  
  m_portFd = open(port, O_RDWR | O_NOCTTY| O_NONBLOCK);
  //m_portFd = open(port, O_RDWR | O_NOCTTY);
  if(m_portFd < 0){
    cerr << "Cannot open " << port << ": " << flush;
    perror("Error");
    return TRobot::ERROR_OPEN_PORT;
  }

  memset(&termCfg, 0, sizeof(termCfg));

  cfmakeraw(&termCfg);
  termCfg.c_iflag |= IGNPAR; // パリティエラーを無視
  termCfg.c_iflag &= ~(IXON | IXOFF | INPCK); // フロー制御なし，パリティチェックなし
  // 受信有効, データビット数 8bit
  termCfg.c_cflag |= (CREAD| CS8);
  // ストップビット 1 bit, CTS/RTS(ハードウェア)フロー制御無効
  termCfg.c_cflag &= ~(CSTOPB | CRTSCTS);
  termCfg.c_lflag &= ~(PARENB); // パリティチェックなし
  termCfg.c_lflag &= ~(ECHO | ICANON); // エコー，行編集を無効化
  termCfg.c_cc[VMIN] = 0; // non-blocking
  termCfg.c_cc[VTIME] = 0;


  cfsetispeed(&termCfg, B115200);
  cfsetospeed(&termCfg, B115200);
  retVal = tcsetattr(m_portFd, TCSANOW, &termCfg);
  if(retVal == -1){
    perror("tcsetattr():");
  }
  retVal = tcflush(m_portFd, TCIOFLUSH);
  if(retVal == -1){
    perror("tcflush():");
  }
  
  retVal = pthread_create(&m_recvThrId, NULL,
                          TRobot::SerialIOLinux::recvThreadFuncWrapper, this);
  if(retVal != 0){
    printPosixError("Failed to create a receiver thread", retVal);
    close(m_portFd);
    return TRobot::ERROR_CREATE_RECV_THREAD;
  }
  cout << "Created a receiver thread" << endl;
  
  retVal = pthread_create(&m_sendThrId, NULL,
                          TRobot::SerialIOLinux::sendThreadFuncWrapper, this);
  if(retVal != 0){
    printPosixError("Failed to create a sender thread", retVal);
    close(m_portFd);
    return TRobot::ERROR_CREATE_SEND_THREAD;
  }
  cout << "Created a sender thread" << endl;

  m_isRunning = true;
  return TRobot::SUCCESS;
}

////////////////////////////////////////////////////////////////////////////////
/*!
 *  @brief 通信を終了する
 *
 *  @copybrief
 */
TRobot::RetCode TRobot::SerialIOLinux::stop()
{
  requestStop();
    
  // 受信スレッドの終了を待つ
  joinThreadErrCheck(&m_recvThrId, "receiver");
  joinThreadErrCheck(&m_sendThrId, "sender");
  
  close(m_portFd);

  m_stopRequested = false;
  m_isRunning = false;

  // 次にstart()が呼ばれるときに，残っているデータのせいで遅延が生じないようにする
  clearSensorDataQueue();
  clearCmdPairQueue();
  return TRobot::SUCCESS;
}

////////////////////////////////////////////////////////////////////////////////
/*!
 *  @brief 通信スレッドの停止要求の有無を確認する
 *
 *  @copybrief
 */
bool TRobot::SerialIOLinux::stopRequested()
{
  bool req;
  pthread_mutex_lock(&m_stopReqLock);
  req = m_stopRequested;
  pthread_mutex_unlock(&m_stopReqLock);
  return req;
}

////////////////////////////////////////////////////////////////////////////////
/*!
 *  @brief 通信スレッドの停止を要求する
 *
 *  @copybrief
 */
void TRobot::SerialIOLinux::requestStop()
{
  pthread_mutex_lock(&m_stopReqLock);
  m_stopRequested = true;
  pthread_mutex_unlock(&m_stopReqLock);
}



////////////////////////////////////////////////////////////////////////////////
/*!
 *  @brief 指定された分のデータを全て受信する
 *
 *  @copybrief
 *
 *  @param fd  通信ポートのファイルディスクリプタ
 *  @param buf 受信データを格納するバッファ
 *  @param numBytes 受信するデータのサイズ [bytes]
 */
TRobot::RetCode TRobot::SerialIOLinux::receiveExactNumBytes(int fd, unsigned char *buf, size_t numBytes)
{
  size_t nrecv = 0;
  ssize_t retVal = 0;
  fd_set fdSet;
  int selectRes = 0;
  FD_ZERO(&fdSet);
    
  while(nrecv < numBytes){
    // select() は timeout を変更しうるので呼び出し前に初期化が必要
    struct timeval timeout = {0, 10*TRobot::CTRL_PERIOD};
    FD_ZERO(&fdSet);
    FD_SET(fd, &fdSet);
    selectRes = select(fd + 1, &fdSet, NULL, NULL, &timeout);
    //selectRes = select(fd + 1, &fdSet, NULL, NULL, NULL);
    if(selectRes < 0){
      perror("Error on select()");
            
    }else if(selectRes == 0){
      std::cerr << "select() timed out" << std::endl;
      continue;
    }
    if(FD_ISSET(fd, &fdSet)){
      retVal = read(fd, buf + nrecv, numBytes - nrecv);
      if(retVal > 0){
        nrecv += retVal;
        std::cerr << "received " << retVal << " bytes." << std::endl;
      }else if(retVal < 0){
        perror("Error");
      }
    }
  }


#if 0// ndef NDEBUG
  using namespace std;
  cerr << "received: ";
  for(size_t i = 0; i < nrecv; i++){
    fprintf(stderr, "%02x ", buf[i]);
  }
  cerr << endl;
#endif
  return TRobot::SUCCESS;  
}

////////////////////////////////////////////////////////////////////////////////
/*!
 *  @brief 指定された分のデータを全て送信する
 *
 *  @copybrief
 *
 *  @param fd  通信ポートのファイルディスクリプタ
 *  @param buf 送信データを格納するバッファ
 *  @param numBytes 送信するデータのサイズ [bytes] 
 */
TRobot::RetCode TRobot::SerialIOLinux::sendExactNumBytes(int fd, unsigned char *buf, size_t numBytes)
{
  size_t nsent = 0;
  ssize_t retVal = 0;
  fd_set fdSet;
  FD_ZERO(&fdSet);

  while(nsent < numBytes){
    FD_SET(fd, &fdSet);
    select(fd + 1, NULL, &fdSet, NULL, NULL);
    
    if(FD_ISSET(fd, &fdSet)){
      retVal = write(m_portFd, buf + nsent, numBytes - nsent);
      if(retVal > 0){
        nsent += retVal;
      }else if(retVal < 0){
        perror("Error");
      }
    }
  }

#ifndef NDEBUG
  using namespace std;
  cerr << "sent: "; 
  for(size_t i = 0; i < nsent; i++){
    fprintf(stderr, "%02x ", buf[i]);
  }
  cerr << endl;
#endif
  return TRobot::SUCCESS;;
}

////////////////////////////////////////////////////////////////////////////////
/*!
 *
 *  次のヘッダが見つかるまで読み込む
 *
 *  ポーリングでなく一方的なストリーミングのため，
 *  読み込んだバッファの先頭がデータのヘッダとは限らない
 *
 *  @param buf    [out]  読み込み用のバッファ
 *  @param bufLen [in]   buf の長さ [bytes]
 */
void TRobot::SerialIOLinux::seekToNextHeader(unsigned char *buf,
                                             size_t bufLen)
{
  using namespace std;
  const unsigned char *headPos = NULL;
  unsigned char *bufLast = &buf[bufLen];
  size_t remDataLen = 0;
  const useconds_t WAIT_DURATION = 8000; // [usec]
  int numSeekRetry = 0;
    
  while (true) { // ヘッダが見つかるまで
    receiveExactNumBytes(m_portFd, buf, BASE_SENSOR_DATA_LEN);
    headPos = searchForValidHeader(buf, bufLast);
    if (headPos != bufLast) {
      break;
    }
    DBGOUT(<< "numSeekRetry == " << numSeekRetry++ << endl);
    usleep(WAIT_DURATION);
  }
  remDataLen = headPos - buf; // ヘッダまでの残りのデータ長
  // cout << "remDataLen == " << remDataLen << endl;

  // remDataLen == 0 であれば buf の先頭にヘッダがある
  if (remDataLen != 0) {
    // 次のヘッダの直前まで読み込んで捨てる
    receiveExactNumBytes(m_portFd, buf, remDataLen);
    receiveExactNumBytes(m_portFd, buf, BASE_SENSOR_DATA_LEN);
    usleep(WAIT_DURATION);
  }
}

////////////////////////////////////////////////////////////////////////////////
/*!
 *  @brief 受信スレッドのルーチンを扱うためのラッパ関数
 *
 *  @copybrief
 *
 *  @pre SerialIOLinux のインスタンスが1個のみ存在する
 *       (現状，この関数は再入可能(reentrant) でない) 
 */
void *TRobot::SerialIOLinux::recvThreadFuncWrapper(void *arg)
{
  TRobot::SerialIOLinux *self;
  void *retVal = (void *)0;
  if (arg != NULL) {
    self = (TRobot::SerialIOLinux*)(arg);
    retVal = self->recvThreadFunc();
  }
  return retVal;
}


////////////////////////////////////////////////////////////////////////////////
/*!
 *  @brief 
 *
 *  @copybrief
 */
void *TRobot::SerialIOLinux::sendThreadFuncWrapper(void *arg)
{
  TRobot::SerialIOLinux *self;
  void *retVal = (void *)0;
  if (arg != NULL) {
    self = (TRobot::SerialIOLinux*)(arg);
    retVal = self->sendThreadFunc();
  }
  return retVal;
}


////////////////////////////////////////////////////////////////////////////////
/*!
 *  @brief 
 *
 *  @copybrief
 */
void *TRobot::SerialIOLinux::sendThreadFunc()
{
  using namespace std;
  static unsigned char cmd[TRobot::ROT_VEL_CMD_LEN];
  RetCode retCode;
  const useconds_t WAIT_DURATION = 8000; // TRobot::CTRL_PERIOD/2; //8000; // [usec]
  const int THREAD_EXIT_SUCCESS = 0;
  RotVelCmdPair cmdPair;
  bool firstSend = true;
  
  while (!stopRequested()) {
    if (cmdPairQueueEmpty()) {
      usleep(WAIT_DURATION);
      continue;
    }
    popCmdPair(cmdPair);
    
    retCode = buildCmdSeqSetRotVel(cmdPair.left(), cmdPair.right(), cmd);
    if (retCode != TRobot::SUCCESS) {
      cerr << "Failed to build a command sequence" << endl;
      break;
    }

    sendExactNumBytes(m_portFd, cmd, TRobot::ROT_VEL_CMD_LEN);

    if (firstSend) {
      pthread_cond_signal(&m_condFirstSendDone);
      firstSend = false;
    }

    usleep(WAIT_DURATION);
  }

  return (void *)THREAD_EXIT_SUCCESS;
}

////////////////////////////////////////////////////////////////////////////////
/*!
 *  @brief 受信スレッドのメインルーチン
 *
 *  @copybrief
 */
void *TRobot::SerialIOLinux::recvThreadFunc()
{
  using namespace TRobot;
  using namespace std;
  unsigned char buf[BASE_SENSOR_DATA_LEN];
  BaseSensorData sensorData;
  const useconds_t WAIT_DURATION = 8000; // [usec]
  const int THREAD_EXIT_SUCCESS = 0;
  pthread_mutex_t condMutex; // 送信スレッドとの待ち合わせのためのmutex
    
  DBGOUT(<< "recv thread starting..." << endl);
  
  initMutexErrCheck(&condMutex);
  {
    pthread_mutex_lock(&condMutex);
    // 最初のコマンド送信を待つ（デッドロックを防ぐため）
    pthread_cond_wait(&m_condFirstSendDone, &condMutex);
    pthread_mutex_unlock(&condMutex);
  }
  while (!stopRequested()) {
    seekToNextHeader(buf, BASE_SENSOR_DATA_LEN);
    decodeBaseSensorData(buf, sensorData);
    pushSensorData(sensorData);
        
    usleep(WAIT_DURATION);
  }
  destroyMutexErrCheck(&condMutex);
  return (void *)THREAD_EXIT_SUCCESS;
}


////////////////////////////////////////////////////////////////////////////////
/*!
 *  @brief 左右の車輪の回転速度の目標値を送信する
 *
 *  @copybrief
 *
 *  @param leftRotVel  左車輪の回転速度指令値
 *  @param rightRotVel 右車輪の回転速度指令値
 */
TRobot::RetCode TRobot::SerialIOLinux::sendWheelRotVel(int leftRotVel, int rightRotVel)
{
  pushCmdPair(RotVelCmdPair(leftRotVel, rightRotVel));
  return TRobot::SUCCESS;  
}

////////////////////////////////////////////////////////////////////////////////
/*!
 *  @brief センサデータを受信する
 *
 *  @copybrief
 *
 *  @param sensorData [出力] センサデータ
 */
TRobot::RetCode TRobot::SerialIOLinux::receiveSensorData(TRobot::BaseSensorData& sensorData)
{
  static const TRobot::BaseSensorData NIL_SENSOR_DATA;

  if (sensorDataQueueEmpty()) {
    //return TRobot::ERROR_NO_DATA;
    return TRobot::ERROR_QUEUE_EMPTY;
    // sensorData = NIL_SENSOR_DATA;
  }
  popSensorData(sensorData);
  return TRobot::SUCCESS;  
}



////////////////////////////////////////////////////////////////////////////////
/*!
 *  @brief センサデータをキューに入れる
 *
 *  @copybrief
 *
 *  スレッドセーフ
 */
void TRobot::SerialIOLinux::pushSensorData(const BaseSensorData& sensorData)
{
  pthread_mutex_lock(&m_senDataQLock);
  m_sensorDataQueue.push_back(sensorData);
  pthread_mutex_unlock(&m_senDataQLock);  
}

////////////////////////////////////////////////////////////////////////////////
/*!
 *  @brief センサデータをキューから取り出す
 *
 *  @copybrief
 *
 *  スレッドセーフ
 */
void TRobot::SerialIOLinux::popSensorData(BaseSensorData& sensorData)
{
  pthread_mutex_lock(&m_senDataQLock);
  sensorData = m_sensorDataQueue.front();
  m_sensorDataQueue.pop_front();
  pthread_mutex_unlock(&m_senDataQLock);  
}

////////////////////////////////////////////////////////////////////////////////
/*!
 *  @brief センサデータのキューが空か否かを確認する
 *
 *  @copybrief
 * 
 *  スレッドセーフ
 *
 *  @return 空であれば true，空でなければ false
 */
bool TRobot::SerialIOLinux::sensorDataQueueEmpty()
{
  bool empty = false;
  pthread_mutex_lock(&m_senDataQLock);
  empty = m_sensorDataQueue.empty();
  pthread_mutex_unlock(&m_senDataQLock);
  return empty;
}

////////////////////////////////////////////////////////////////////////////////
/*!
 *  @brief センサデータのキューを空にする
 *
 *  @copybrief
 * 
 *  スレッドセーフ
 */
void TRobot::SerialIOLinux::clearSensorDataQueue()
{
  pthread_mutex_lock(&m_senDataQLock);
  m_sensorDataQueue.clear();
  pthread_mutex_unlock(&m_senDataQLock);  
}

////////////////////////////////////////////////////////////////////////////////
/*!
 *  @brief 指令値をキューに入れる
 *
 *  @copybrief
 *
 *  スレッドセーフ
 */
void TRobot::SerialIOLinux::pushCmdPair(const RotVelCmdPair& cmdPair)
{
  pthread_mutex_lock(&m_rotVelCmdQLock);
  m_rotVelCmdQueue.push_back(cmdPair);
  pthread_mutex_unlock(&m_rotVelCmdQLock);
}


////////////////////////////////////////////////////////////////////////////////
/*!
 *  @brief 指令値をキューから取り出す
 *
 *  @copybrief
 *
 *  スレッドセーフ
 */
void TRobot::SerialIOLinux::popCmdPair(RotVelCmdPair& cmdPair)
{
  pthread_mutex_lock(&m_rotVelCmdQLock);
  cmdPair = m_rotVelCmdQueue.front();
  m_rotVelCmdQueue.pop_front();    
  pthread_mutex_unlock(&m_rotVelCmdQLock);
}


////////////////////////////////////////////////////////////////////////////////
/*!
 *  @brief 指令値のキューが空か否かを確認する
 *
 *  @copybrief
 * 
 *  スレッドセーフ
 *
 *  @return 空であれば true，空でなければ false
 */
bool TRobot::SerialIOLinux::cmdPairQueueEmpty()
{
  bool empty = false;
  pthread_mutex_lock(&m_rotVelCmdQLock);
  empty = m_rotVelCmdQueue.empty();
  pthread_mutex_unlock(&m_rotVelCmdQLock);
  return empty;  
}


////////////////////////////////////////////////////////////////////////////////
/*!
 *  @brief 指令値のキューを空にする
 *
 *  @copybrief
 * 
 *  スレッドセーフ
 */
void TRobot::SerialIOLinux::clearCmdPairQueue()
{
  pthread_mutex_lock(&m_rotVelCmdQLock);
  m_rotVelCmdQueue.clear();
  pthread_mutex_unlock(&m_rotVelCmdQLock);  
}


////////////////////////////////////////////////////////////////////////////////
/*!
 *
 *  @return 通信中であれば true，そうでなければ false
 */
bool TRobot::SerialIOLinux::isRunning()
{
  return m_isRunning;
}
